// SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK, http://msk.desy.de
// SPDX-License-Identifier: LGPL-3.0-or-later
#pragma once

#include <tec/TecChimeraInterface.h>
#include <tec/TecController.h>
#include <tec/TecFrame.h>

#include <ChimeraTK/ApplicationCore/ApplicationCore.h>

namespace ctk = ChimeraTK;

struct TecModule : ctk::ApplicationModule {
  TecModule(ctk::ModuleGroup* owner, const std::string& name, const std::string& description,
      const std::unordered_set<std::string>& tags = {});

  TecModule() = default;
  TecModule(TecModule&& other) noexcept;

  ~TecModule() override = default;

  ctk::ScalarPushInput<uint64_t> periodicReadTrigger{this, "/PeriodicRead/tick", "", ""};

  // access to TEC library
  ctk::ScalarOutput<ctk::Boolean> triggerCommands{this, "triggerCommands", "", ""};
  ctk::ScalarPollInput<ctk::Boolean> responseStatus{this, "responseStatus", "", ""};
  ctk::ArrayPollInput<uint32_t> commandMemoryRaw_rd{this, "commandMemoryRaw_rd", "", 256, ""};
  ctk::ArrayOutput<uint32_t> commandMemoryRaw_wr{this, "commandMemoryRaw_wr", "", 256, ""};

  ctk::ScalarPollInput<uint32_t> tecAddress{
      this, "tecAddress", "", "The address of the TEC device on the serial interface"};

  struct ErrorLog : ctk::VariableGroup {
    using ctk::VariableGroup::VariableGroup;
    ctk::ScalarOutput<std::string> errorMessage{
        this, "errorMessage", "", "Message in case of errors. Need to be reset via errorReset."};
    ctk::ScalarOutput<ctk::Boolean> errorStatus{
        this, "errorStatus", "", "Boolean flag whether there is an error. Need to be reset via errorReset."};
    ctk::ScalarPushInput<int32_t> errorReset{this, "errorReset", "", "Clear errorMessage and errorStatus."};

    void append(const std::string& e);
    void reset();
    void write();
    bool dirty{false};
  } errorlog{this, ".", ""};

  ctk::ScalarPushInputWB<uint32_t> fpgaTimeout{
      this, "fpgaTimeout", "ms", "Timeout for communication with FPGA firmware."};

  ctk::ScalarPollInput<ctk::Boolean> enableDebug{this, "enableDebug", "", "Enable debug output to console"};
  ctk::ScalarPollInput<ctk::Boolean> enableHeartbeat{
      this, "enableHeartbeat", "", "Enable periodic check of communication with TEC"};

  ctk::ScalarPushInput<int> resetDevice{this, "resetTecDevice", "", "Send reset command to the TEC device"};

  void mainLoop() override;
  void readPeriodic();
  void readGroup(std::list<TecFrame::TecParameter>& group);
  void writeGroup(std::list<TecFrame::TecParameter>& group);
  void writeGroupToFlash(std::list<TecFrame::TecParameter>& group);
  void writeGroupInternal(std::list<TecFrame::TecParameter>& group, std::vector<TecFrame>& writeRequests);

  /**
   * @brief Reset the TEC device
   */
  void doResetDevice();

  /** Perform the actual communication with the TEC device. The list of frames passed to this function is sent to the
   *  device, returned is a list of frames with the replies from the device. If the list exceeds a certain limit,
   *  multiple transfers will be initiated.
   *
   *  Method returns a list of pointers to the processed responses. All requests that fail due to a timeout is marked as
   *  a nullptr in the return list.
   *  */
  std::vector<std::shared_ptr<TecFrame>> doTecCommunication(std::vector<TecFrame>& requests);

  template<typename T>
  void checkAndUpdateReadResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor);

  template<typename T>
  void checkAndUpdateWriteResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor);

  std::list<std::pair<ChimeraTK::ScalarPollInput<int>, TecFrame::TecParameter>> enablePeriodicReadList;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarPushInputWB<int>> writeIntMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarPushInputWB<float>> writeFloatMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarOutput<int>> readIntMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarOutput<float>> readFloatMap;
  std::map<std::string, ChimeraTK::VoidInput> groupCommandMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> writeGroupMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> writeGroupToFlashMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> readGroupMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> setInputFromOutputGroupMap;
  // std::unordered_set<ChimeraTK::TransferElementID> writeBackAccessors;

  void setInputFromOutputGroup(std::list<TecFrame::TecParameter>& group);
  void createReadWriteMaps();
  void createTransferIdGroupMaps();
  void createCommandGroupMaps();
  void createEnablePeriodicList();
  void createWriteBackAccessorsUnorderedSet();

  std::shared_ptr<TecChimeraInterface> tecIface;
  TecController tec{tecIface};
};
