// SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK, http://msk.desy.de
// SPDX-License-Identifier: LGPL-3.0-or-later

// Define a name for the test module.
#define BOOST_TEST_MODULE testAllParameters

// Only after defining the name include the unit test header.
#include "Server.h"
#include "TecDummy.h"
#include "TecParameters.h"

#include <ChimeraTK/ApplicationCore/TestFacility.h>
#include <ChimeraTK/BackendFactory.h>

#include <boost/test/included/unit_test.hpp>

#include <random>

using namespace boost::unit_test_framework;

struct Fixture_t {
  Server theServer{"tec_dummy"};
  struct LocalTestFacility : public ChimeraTK::TestFacility {
    using ChimeraTK::TestFacility::TestFacility;
    boost::shared_ptr<ChimeraTK::ControlSystemPVManager> getPvManager() { return pvManager; }
  } testFacility{theServer};
  boost::shared_ptr<TecDummy> dummy{
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend("ActualDevice"))};
};

/// TestSuite for the server, adapt name
BOOST_AUTO_TEST_SUITE(AllParametersTestSuite)

/**********************************************************************************************************************/

BOOST_FIXTURE_TEST_CASE(testControlSystemVariables, Fixture_t) {
  std::unordered_set<std::string> readGroups;
  std::unordered_set<std::string> writeGroups;
  std::unordered_set<std::string> writeToFlashGroups;

  testFacility.runApplication();

  auto pvManager = testFacility.getPvManager();

  for(auto& params : descriptors) {
    if(not params.readGroup.empty()) readGroups.insert(params.readGroup);
    if(not params.writeGroup.empty()) writeGroups.insert(params.writeGroup);
    if(not params.writeToFlashGroup.empty()) writeToFlashGroups.insert(params.writeToFlashGroup);

    // This variable has to exist
    BOOST_CHECK_NO_THROW(pvManager->getProcessVariable("/TecModule0/" + params.pvName + "_rd"));
    auto pv = pvManager->getProcessVariable("/TecModule0/" + params.pvName + "_rd");
    BOOST_CHECK(pv->isReadOnly());
    if(params.valueType == ValueType::Int) {
      BOOST_CHECK_MESSAGE(pv->getValueType() == typeid(int), "Check if _rd valueType is integer");
    }
    else if(params.valueType == ValueType::Float) {
      BOOST_CHECK_MESSAGE(pv->getValueType() == typeid(float), "Check if _rd valueType is float");
    }
    else {
      BOOST_ASSERT(false);
    }

    // if this variable exists, it has to be
    try {
      pv = pvManager->getProcessVariable("/TecModule0/" + params.pvName + "_enaPeriodicRead");
      BOOST_CHECK(pv->isWriteable());
      BOOST_CHECK_MESSAGE(pv->getValueType() == typeid(int32_t), "Check if _enaPeriodicRead valueType is int32_t");
      BOOST_CHECK(params.direction == Direction::Read);
    }
    catch(ChimeraTK::logic_error&) {
      // If this variable does not exist, the parameter has to be read-only
      BOOST_CHECK(params.direction == Direction::ReadWrite);
    }

    try {
      pv = pvManager->getProcessVariable("/TecModule0/" + params.pvName + "_wr");
      BOOST_CHECK(params.direction == Direction::ReadWrite);
      BOOST_CHECK(pv->isWriteable());
      if(params.valueType == ValueType::Int) {
        BOOST_CHECK_MESSAGE(pv->getValueType() == typeid(int), "Check if _wr valueType is integer");
      }
      else if(params.valueType == ValueType::Float) {
        BOOST_CHECK_MESSAGE(pv->getValueType() == typeid(float), "Check if _wr valueType is float");
      }
      else {
        BOOST_ASSERT(false);
      }
    }
    catch(ChimeraTK::logic_error&) {
      // If this variable does not exist, the parameter has to be read-only
      BOOST_CHECK(params.direction == Direction::Read);
    }
  }

  // Check if the group accessors exist
  for(const auto& group : {readGroups, writeGroups, writeToFlashGroups}) {
    for(const auto& entry : group) {
      BOOST_REQUIRE_NO_THROW(pvManager->getProcessVariable("/TecModule0/" + entry + "_wr"));
      auto pv = pvManager->getProcessVariable("/TecModule0/" + entry + "_wr");
      BOOST_CHECK(pv->isWriteable());
    }
  }
}
/**********************************************************************************************************************/

BOOST_FIXTURE_TEST_CASE(testGroupValues, Fixture_t) {
  std::map<std::string, std::vector<decltype(descriptors)::iterator>> readGroup;
  std::map<std::string, std::vector<decltype(descriptors)::iterator>> writeGroup;
  std::map<std::string, std::vector<decltype(descriptors)::iterator>> writeToFlashGroup;

  // Prepare a list of random number to use as data for each group
  // The maximum canaries we need is number of descriptors times 3, if all
  // descriptors were in all three groups
  std::unordered_set<unsigned int> canaries;
  std::random_device r;
  while(canaries.size() < descriptors.size() * 3) {
    canaries.insert(r());
  }

  for(auto params = descriptors.begin(); params != descriptors.end(); ++params) {
    if(not params->readGroup.empty()) {
      readGroup[params->readGroup].emplace_back(params);
    }
    if(not params->writeGroup.empty()) {
      writeGroup[params->writeGroup].emplace_back(params);
    }
    if(not params->writeToFlashGroup.empty()) {
      writeToFlashGroup[params->writeToFlashGroup].emplace_back(params);
    }
  }

  testFacility.runApplication();

  // Check reading - Set the canary value in the dummy, trigger read-out of group
  // And see that all members of the group are now set to the canary value
  for(auto& group : readGroup) {
    // Get a unique marker
    auto canary = *canaries.begin();
    canaries.erase(canaries.begin());

    // Check that all CS variables are not containing the canary value
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        BOOST_CHECK_NE(
            testFacility.readScalar<float>("/TecModule0/" + param->pvName + "_rd"), static_cast<float>(canary));
      }
      else if(param->valueType == ValueType::Int) {
        BOOST_CHECK_NE(testFacility.readScalar<int>("/TecModule0/" + param->pvName + "_rd"), canary);
      }
    }

    // Fill all the parameters in the dummy device that are contained in the group
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        dummy->set(param->value, static_cast<float>(canary));
      }
      else if(param->valueType == ValueType::Int) {
        dummy->set(param->value, static_cast<int>(canary));
      }
      else {
        BOOST_REQUIRE(false);
      }
    }

    std::cout << "Testing read group " << group.first << std::endl;

    testFacility.getVoid("/TecModule0/" + group.first + "_wr").write();
    testFacility.stepApplication();

    // Check that now all CS variables actually contain the canary value
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        BOOST_CHECK_CLOSE(
            testFacility.readScalar<float>("/TecModule0/" + param->pvName + "_rd"), static_cast<float>(canary), 1e5F);
      }
      else if(param->valueType == ValueType::Int) {
        BOOST_CHECK_EQUAL(testFacility.readScalar<int>("/TecModule0/" + param->pvName + "_rd"), canary);
      }
    }
  }

  // Check normal writing
  for(auto& group : writeGroup) {
    // Get a unique marker
    auto canary = *canaries.begin();
    canaries.erase(canaries.begin());

    // Write inverse pattern into all parameter registers in the dummy and set the
    // Last write to flash flag on them
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        dummy->set(param->value, static_cast<float>(~canary));
      }
      else if(param->valueType == ValueType::Int) {
        dummy->set(param->value, static_cast<int>(~canary));
      }
      else {
        BOOST_REQUIRE(false);
      }
      dummy->setLastWriteToFlash(param->value, true);
    }

    // Set saving to flash on to check if the group properly disables it
    dummy->set(TecFrame::ParameterSystemFlashSaveOff, 0);

    // Set all CS variables to our canary value
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        testFacility.writeScalar<float>("/TecModule0/" + param->pvName + "_wr", static_cast<float>(canary));
      }
      else if(param->valueType == ValueType::Int) {
        testFacility.writeScalar<int>("/TecModule0/" + param->pvName + "_wr", static_cast<int>(canary));
      }
    }

    std::cout << "Testing write group " << group.first << std::endl;

    testFacility.getVoid("/TecModule0/" + group.first + "_wr").write();
    testFacility.stepApplication();

    // Check in dummy that the requested parameters now contain the canary value
    // and the last write to flash flag was cleared
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        BOOST_CHECK_CLOSE(dummy->getFloat(param->value), static_cast<float>(canary), 1e5F);
      }
      else if(param->valueType == ValueType::Int) {
        BOOST_CHECK_EQUAL(dummy->getInt(param->value), canary);
      }
      else {
        BOOST_REQUIRE(false);
      }
      BOOST_REQUIRE(not dummy->getLastWriteToFlash(param->value));
    }
  }

  // Check writing to flash
  for(auto& group : writeToFlashGroup) {
    // Get a unique marker
    auto canary = *canaries.begin();
    canaries.erase(canaries.begin());

    // Check in dummy that the requested parameters do not already contain the canary
    // and clear the writeToFlash flag for those parameters
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        dummy->set(param->value, static_cast<float>(~canary));
      }
      else if(param->valueType == ValueType::Int) {
        dummy->set(param->value, static_cast<int>(~canary));
      }
      else {
        BOOST_REQUIRE(false);
      }
      dummy->setLastWriteToFlash(param->value, false);
    }

    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        testFacility.writeScalar<float>("/TecModule0/" + param->pvName + "_wr", static_cast<float>(canary));
      }
      else if(param->valueType == ValueType::Int) {
        testFacility.writeScalar<int>("/TecModule0/" + param->pvName + "_wr", static_cast<int>(canary));
      }
    }

    std::cout << "Testing write group " << group.first << std::endl;

    // Set saving to flash off to check if the group properly enables it
    dummy->set(TecFrame::ParameterSystemFlashSaveOff, 1);

    testFacility.getVoid("/TecModule0/" + group.first + "_wr").write();
    testFacility.stepApplication();

    // Fill all the
    for(auto& param : group.second) {
      if(param->valueType == ValueType::Float) {
        BOOST_CHECK_CLOSE(dummy->getFloat(param->value), static_cast<float>(canary), 1e5F);
      }
      else if(param->valueType == ValueType::Int) {
        BOOST_CHECK_EQUAL(dummy->getInt(param->value), static_cast<int>(canary));
      }
      else {
        BOOST_REQUIRE(false);
      }
      BOOST_REQUIRE(dummy->getLastWriteToFlash(param->value));
    }
  }
}

/**********************************************************************************************************************/

BOOST_AUTO_TEST_SUITE_END()
