// SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK, http://msk.desy.de
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "TecDummy.h"

#include <arpa/inet.h>
#include <tec/TecFrame.h>

#include <boost/crc.hpp>

#include <sstream>

// Read-only tec variables as per tec communication specification document
static std::vector<TecFrame::TecParameter> READ_ONLY_VARS = {
    // Parameters 100-999 are read-only, except 108 (Enable/disable save to flash)
    TecFrame::TecParameter::DeviceType,
    TecFrame::TecParameter::HardwareVersion,
    TecFrame::TecParameter::SerialNumber,
    TecFrame::TecParameter::FirmwareVersion,
    TecFrame::TecParameter::DeviceStatus,
    TecFrame::TecParameter::ErrorNumber,
    TecFrame::TecParameter::ErrorInstance,
    TecFrame::TecParameter::ErrorParameter,
    TecFrame::TecParameter::ParameterSystemFlashStatus,

    // All monitoring parmaters are read-only
    TecFrame::TecParameter::MonObjectTemperature,
    TecFrame::TecParameter::MonSinkTemperature,
    TecFrame::TecParameter::MonTargetObjectTemperature,
    TecFrame::TecParameter::MonRampNominalObjectTemperature,
    TecFrame::TecParameter::MonThermalPowerModelCurrent,
    TecFrame::TecParameter::MonActualOutputCurrent,
    TecFrame::TecParameter::MonActualOutputVoltage,
    TecFrame::TecParameter::MonPIDLowerLimitation,
    TecFrame::TecParameter::MonPIDUpperLimitation,
    TecFrame::TecParameter::MonPIDControlVariable,
    TecFrame::TecParameter::MonObjectSensorRawADCValue,
    TecFrame::TecParameter::MonSinkSensorRawADCValue,
    TecFrame::TecParameter::MonObjectSensorResistance,
    TecFrame::TecParameter::MonSinkSensorResitance,
    TecFrame::TecParameter::MonSinkSensorTemperature,
    TecFrame::TecParameter::MonObjectSensorTemperature,
    TecFrame::TecParameter::MonObjectSensorType, // and 1046
    TecFrame::TecParameter::MonFirmwareVersion,
    TecFrame::TecParameter::MonFirmwareBuildNumber,
    TecFrame::TecParameter::MonHardwareVersion,
    TecFrame::TecParameter::MonSerialNumber,
    TecFrame::TecParameter::MonDriverInputVoltage,
    TecFrame::TecParameter::MonMedVInternalSupply,
    TecFrame::TecParameter::Mon33VInternalSupply,
    TecFrame::TecParameter::MonBasePlateTemperature,
    TecFrame::TecParameter::MonErrorNumber,
    TecFrame::TecParameter::MonErrorInstance,
    TecFrame::TecParameter::MonErrorParameter,
    TecFrame::TecParameter::MonParallelActualOutputCurrent,
    TecFrame::TecParameter::MonDriverStatus,
    TecFrame::TecParameter::MonParameterSystemFlashStatus,
    TecFrame::TecParameter::MonFanRelativeCoolingPower,
    TecFrame::TecParameter::MonFanNominalFanSpeed,
    TecFrame::TecParameter::MonFanActualFanSpeed,
    TecFrame::TecParameter::MonFanActualPwmLevel,
    TecFrame::TecParameter::MonTemperatureIsStable,
};

#define TEC_MAGIC_TIMEOUT 0xfaceb00b

TecDummy::BackendRegisterer TecDummy::backendRegisterer;

/*********************************************************************************************************************/

TecDummy::BackendRegisterer::BackendRegisterer() {
  std::cout << "TecDummy::BackendRegisterer: registering backend type TecDummy" << std::endl;
  ChimeraTK::BackendFactory::getInstance().registerBackendType("TecDummy", &TecDummy::createInstance);
}

/*********************************************************************************************************************/

void TecDummy::read(uint64_t bar, uint64_t address, int32_t* data, size_t sizeInBytes) {
  DummyBackend::read(bar, address, data, sizeInBytes);

  if(bar != TecDummy_bar || address != TecDummy_address_loopDone) return;
  if(reg_loopStart != 1) return;
  if(loopDoneDelay) {
    loopDoneDelay--;
    return;
  }

  // convert command into string
  // FIXME: this can be done better if the DummyRegisterRawAccessor would allow us to use the raw pointer to this...
  std::string commandString(reg_command.getNumberOfElements() * 4UL, 0);
  for(auto i = 0U; i < reg_command.getNumberOfElements(); ++i) {
    uint32_t word = ntohl(reg_command[i]);
    std::memcpy(&(commandString[4UL * i]), &word, 4);
  }

  // process frame by frame
  constexpr size_t nFrames = 16;
  constexpr size_t frameLength = 16 * 4UL;
  std::string reply;
  for(size_t frame = 0; frame < nFrames; ++frame) {
    size_t start = frame * frameLength + 8;
    auto frameData = commandString.substr(start, frameLength);

    // remove everything from the 0x0D character, because only 0x00 following
    auto endOfFrame = frameData.find_first_of('\r');
    if(endOfFrame == std::string::npos) {
      reply += std::string(frameLength, '\0');
      continue;
    }
    frameData.erase(endOfFrame);

    std::string frameReply;
    if(not simulateSerialTimeout) {
      // perform data processing
      frameReply = processFrame(frameData);
    }
    else {
      // In case of timeout, the marker is written to the start of the frame
      // The byte order is swapped here to compensate swapping of the reg_command,
      // this is necessary, because the TEC lib checks the timeout marker before
      // changing the byte order of the received frame.
      const uint32_t timeoutMarker{htonl(TEC_MAGIC_TIMEOUT)};
      frameReply = std::string(sizeof(uint32_t), '\0');
      std::memcpy(&(frameReply[0]), &timeoutMarker, sizeof(uint32_t));
    }

    // append to answer
    frameReply += "\r";
    frameReply += std::string(frameLength - frameReply.size(), '\0');
    reply += frameReply;
  }
  assert(reply.size() == nFrames * frameLength);

  // convert reply back into byte stream
  for(size_t i = 0; i < reply.length() / 4; ++i) {
    uint32_t word;
    std::memcpy(&word, &(reply[4 * i]), 4);
    reg_command[static_cast<unsigned int>(i)] = htonl(word);
  }
  *data = 1;
}

/*********************************************************************************************************************/

std::map<TecDummy::TecFramePart, std::string> TecDummy::splitFrame(const std::string& command) {
  std::map<TecDummy::TecFramePart, std::string> result{};
  unsigned int offset = 0;
  // Request or response marker
  result[TecFramePart::MARKER] = command.substr(offset, 1);
  offset++;

  // TEC address
  result[TecFramePart::ADDRESS] = command.substr(offset, 2);
  offset += 2;

  // Sequence ID
  result[TecFramePart::SEQUENCE] = command.substr(offset, 4);
  offset += 4;

  // Command
  if(command[offset] == '?') {
    result[TecFramePart::COMMAND] = command.substr(offset, 3);
    offset += 3;
  }
  else {
    result[TecFramePart::COMMAND] = command.substr(offset, 2);
    offset += 2;
  }

  // Reset has the command string as only payload. The CRC will follow
  // straight after
  if(result[TecFramePart::COMMAND] != "RS") {
    // Parameter ID
    result[TecFramePart::PARAMETER_ID] = command.substr(offset, 4);
    offset += 4;

    // Channel ID
    result[TecFramePart::CHANNEL] = command.substr(offset, 2);
    offset += 2;
  }

  // payload
  auto residual_length = command.length() - offset - 4;

  if(residual_length > 0) {
    result[TecFramePart::PAYLOAD] = command.substr(offset, residual_length);
    offset += residual_length;
  }

  // CRC
  result[TecFramePart::CRC] = command.substr(offset, 4);

  return result;
}

std::string TecDummy::getCrc(std::string& data) const {
  // CRC-16 CCITT, no reflections, no final xor , with an initial value 0 - which boils down to the XMODEM variant of
  // the CRC16-CCITT.
  boost::crc_xmodem_t crc;

  crc.process_bytes(data.c_str(), data.length());

  if(this->simulateCrcError) {
    crc.process_byte(static_cast<unsigned char>(rand() % std::numeric_limits<uint8_t>::max()));
  }

  std::stringstream ss;
  ss << std::hex << std::setw(4) << std::setfill('0') << std::uppercase << crc.checksum();

  return ss.str();
}

std::string TecDummy::processFrame(const std::string& command) {
  std::cout << "TecDummy::processFrame() command = " << command << std::endl;
  auto parsedFrame = splitFrame(command);
  std::string result;

  // extract parameter ID
  auto parameterId =
      static_cast<TecFrame::TecParameter>(std::stoul("0x" + parsedFrame[TecFramePart::PARAMETER_ID], nullptr, 16));

  if(errorFlags[parameterId] != TecErrorCode::NO_ERROR) {
    // Returning an error has been requested through the dummy interface
    result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];
    result += "+"; // marks the error
    std::stringstream ss;
    ss << std::hex << std::setw(2) << std::setfill('0') << std::uppercase << static_cast<int>(errorFlags[parameterId]);
    result += ss.str();
    result += getCrc(result);
  }
  else if(parsedFrame[TecFramePart::COMMAND] == "VS") {
    if(std::find(READ_ONLY_VARS.begin(), READ_ONLY_VARS.end(), parameterId) != READ_ONLY_VARS.end()) {
      // Write requested to a read-only variable
      result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];
      result += "+"; // marks the error
      std::stringstream ss;
      ss << std::hex << std::setw(2) << std::setfill('0') << std::uppercase
         << static_cast<int>(TecErrorCode::EER_PAR_NOT_WRITABLE);
      result += ss.str();
      result += getCrc(result);
    }
    // Handle store command
    int32_t value = static_cast<int32_t>(std::stol("0x" + parsedFrame[TecFramePart::PAYLOAD], nullptr, 16));

    std::cout << "Storing raw value " << value << "(" << parsedFrame[TecFramePart::PAYLOAD] << ") for parameter "
              << parameterId << std::endl;

    rawValues[parameterId] = value;
    result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];

    // Add CRC of set message as ACK payload
    result += parsedFrame[TecFramePart::CRC];

    result += getCrc(result);

    //
    // mark received parameter as written to flash or not
    auto isLastWriteToFlash = rawValues.at(TecFrame::ParameterSystemFlashSaveOff) == 0;
    lastWriteToFlash[parameterId] = isLastWriteToFlash;
  }
  else if(parsedFrame[TecFramePart::COMMAND] == "?VR") {
    // Handle read command
    result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];
    std::stringstream ss;
    ss << std::hex << std::setw(8) << std::setfill('0') << std::uppercase << rawValues[parameterId];
    result += ss.str();
    result += getCrc(result);
  }

  std::cout << "Returning result " << result << std::endl;

  return result;
}

/*********************************************************************************************************************/
