#!/usr/bin/env python3
# SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK, http://msk.desy.de
# SPDX-License-Identifier: LGPL-3.0-or-later

# Script to generate a DOOCS mapping template for config genrator from the TEC server's variable houshold
# Note: While the generated template accepts arbitrary names for the Tec modules in the target server, this script
# relies on one module being called TecModule0

from lxml import etree
import re

# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
def camel_to_snake(name: str) -> str:
    name = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub(r'([a-z\d])([A-Z])', r'\1_\2', name).lower()

def map_variable(name: str, property_name: str, history: bool = False, zmq : bool = False) -> None:
    # replace ENABLE_PERIODIC_READBACK with PERIODIC_READBACK_ENABLED for backwards compatiblity
    print(f'<property name="TEC.${{moduleNumber}}.{property_name}" source="${{TEC_MODULE_PATHS[moduleNumber]}}/{name}">')
    print(f'  <has_history>{history}</has_history>')
    print(f'  <publish_ZMQ>{zmq}</publish_ZMQ>')
    print('</property>')
    pass

def generate_doocs_variable_mapping() -> None:
    tree = etree.parse("tec_server.xml")
    root = tree.getroot()

    ns = { 'ac' : "https://github.com/ChimeraTK/ApplicationCore" }
    p = etree.XPath(".//ac:directory[@name='TecModule0']/ac:variable"
                    "[starts-with(@name,'mon_') or starts-with(@name, 'ctrl_')]", namespaces=ns)
    variables = p(root)

    print('##mako')
    print('## This template file was autogenerated from the tec server module repository. Do not modify directly')
    print('## It requires the following variables to be set:')
    print('##')
    print('## - TEC_MODULE_PATHS: A list of TEC module number to path in module hierarchy.')
    print('##                     TEC_MODULE_PATHS=["/Sensors/Temperature/Board/Tec", "/Sensors/Box/Temperature/Controller"]')
    print('##')
    print('%for moduleNumber in range(len(TEC_MODULE_PATHS)):')
    print('<!-- Module control variables for this module -->')
    map_variable("errorStatus", "ERROR.STATUS")
    map_variable("errorMessage", "ERROR.MESSAGE")
    map_variable("errorReset", "ERROR.RESET")
    map_variable("enableDebug", "DEBUG.ENA")
    map_variable("enableHeartbeat", "ENABLE_HEARTBEAT.WR")
    map_variable("fpgaTimeout", "FPGA_TIMEOUT.WR")
    map_variable("tecAddress", "BOARD_ADDRESS.WR")

    print('<!-- TEC parameters -->')
    for variable in variables:
        name = variable.attrib["name"]

        property_parts = [x.upper() for x in map(camel_to_snake, name.split("_"))]
        property_name = ".".join(property_parts)
        map_variable(name, property_name, property_parts[-1] == "RD")


    p = etree.XPath(".//ac:directory[@name='TecModule0']/ac:variable"
                    "[starts-with(@name,'readGroup')]", namespaces=ns)

    print("<!-- global read groups -->")
    variables = p(root)
    for variable in variables:
        name = variable.attrib["name"]

        property_parts = [x.upper() for x in map(camel_to_snake, name.split("_"))]
        property_parts.insert(0, "MON")
        property_name = ".".join(property_parts)
        map_variable(name, property_name)

    print("<!-- Global control groups -->")
    map_variable("readWriteGroup1_wr", "CTRL.READ_WRITE_GROUP1.WR")
    map_variable("writeGroup1_wr", "CTRL.WRITE_GROUP1.WR")
    map_variable("writeToFlashGroup1_wr", "CTRL.WRITE_TO_FLASH_GROUP1.WR")
    map_variable("setRbGroup1_wr", "CTRL.SET_RB_GROUP1.WR")

    map_variable("resetTecDevice", "CTRL.RESET_DEVICE.WR")

    p = etree.XPath(".//ac:directory[@name='TecModule0']/ac:variable"
                    "[starts-with(@name,'readCh') or starts-with(@name,'writeCh')]", namespaces=ns)

    print("<!-- per-channel read and write groups -->")
    variables = p(root)
    for variable in variables:
        name = variable.attrib["name"]

        property_parts = [x.upper() for x in map(camel_to_snake, name.split("_"))]
        # Extract channel and make it its own . part
        a = property_parts[0].split('_')
        property_parts.pop(0)
        property_parts.insert(0, a[1])
        a.pop(1)
        property_parts.insert(1, '_'.join(a))
        if property_parts[1].startswith('READ_WRITE') or property_parts[1].startswith('WRITE'):
            property_parts.insert(0, "CTRL")
        else:
            property_parts.insert(0, "MON")
        property_name = ".".join(property_parts)
        map_variable(name, property_name)

    p = etree.XPath(".//ac:directory[@name='TecModule0']/ac:variable"
                    "[starts-with(@name,'setRbCh')]", namespaces=ns)

    print("<!-- per-channel read-back groups -->")
    variables = p(root)
    for variable in variables:
        name = variable.attrib["name"]

        property_parts = [x.upper() for x in map(camel_to_snake, name.split("_"))]
        # Extract channel and make it its own . part
        a = property_parts[0].split('_')
        property_parts.pop(0)
        property_parts.insert(0, a[2])
        a.pop(2)
        property_parts.insert(1, '_'.join(a))
        property_parts.insert(0, "CTRL")
        property_name = ".".join(property_parts)
        map_variable(name, property_name)


    p = etree.XPath(".//ac:directory[@name='TecModule0']/ac:variable"
                    "[starts-with(@name,'writeToFlashCh')]", namespaces=ns)

    print("<!-- per-channel write to flash groups -->")
    variables = p(root)
    for variable in variables:
        name = variable.attrib["name"]

        property_parts = [x.upper() for x in map(camel_to_snake, name.split("_"))]
        # Extract channel and make it its own . part
        a = property_parts[0].split('_')
        property_parts.pop(0)
        property_parts.insert(0, a[3])
        a.pop(3)
        property_parts.insert(1, '_'.join(a))
        property_parts.insert(0, "CTRL")
        property_name = ".".join(property_parts)
        map_variable(name, property_name)


    print('%endfor')


if __name__ == "__main__":
    generate_doocs_variable_mapping()
